from typing import Dict, Text, Any, List, Union, Optional

from rasa_sdk import Tracker
from rasa_sdk.executor import CollectingDispatcher
from rasa_sdk.forms import FormAction


class RestaurantForm(FormAction):
    """Example of a custom form action"""

    def name(self) -> Text:
        """Unique identifier of the form"""

        return "restaurant_form"

    @staticmethod
    def required_slots(tracker: Tracker) -> List[Text]:
        """A list of required slots that the form has to fill"""

        return ["cuisine", "num_people", "outdoor_seating", "preferences", "feedback"]

    def slot_mappings(self) -> Dict[Text, Union[Dict, List[Dict]]]:
        """A dictionary to map required slots to
            - an extracted entity
            - intent: value pairs
            - a whole message
            or a list of them, where a first match will be picked"""

        return {
            "cuisine": self.from_entity(entity="cuisine", not_intent="chitchat"),
            "num_people": [
                self.from_entity(
                    entity="num_people", intent=["inform", "request_restaurant"]
                ),
                self.from_entity(entity="number"),
            ],
            "outdoor_seating": [
                self.from_entity(entity="seating"),
                self.from_intent(intent="affirm", value=True),
                self.from_intent(intent="deny", value=False),
            ],
            "preferences": [
                self.from_intent(intent="deny", value="no additional preferences"),
                self.from_text(not_intent="affirm"),
            ],
            "feedback": [self.from_entity(entity="feedback"), self.from_text()],
        }

    # USED FOR DOCS: do not rename without updating in docs
    @staticmethod
    def cuisine_db() -> List[Text]:
        """Database of supported cuisines"""

        return [
            "caribbean",
            "chinese",
            "french",
            "greek",
            "indian",
            "italian",
            "mexican",
        ]

    @staticmethod
    def is_int(string: Text) -> bool:
        """Check if a string is an integer"""

        try:
            int(string)
            return True
        except ValueError:
            return False

    # USED FOR DOCS: do not rename without updating in docs
    def validate_cuisine(
        self,
        value: Text,
        dispatcher: CollectingDispatcher,
        tracker: Tracker,
        domain: Dict[Text, Any],
    ) -> Dict[Text, Any]:
        """Validate cuisine value."""

        if value.lower() in self.cuisine_db():
            # validation succeeded, set the value of the "cuisine" slot to value
            return {"cuisine": value}
        else:
            dispatcher.utter_message(template="utter_wrong_cuisine")
            # validation failed, set this slot to None, meaning the
            # user will be asked for the slot again
            return {"cuisine": None}

    def validate_num_people(
        self,
        value: Text,
        dispatcher: CollectingDispatcher,
        tracker: Tracker,
        domain: Dict[Text, Any],
    ) -> Dict[Text, Any]:
        """Validate num_people value."""

        if self.is_int(value) and int(value) > 0:
            return {"num_people": value}
        else:
            dispatcher.utter_message(template="utter_wrong_num_people")
            # validation failed, set slot to None
            return {"num_people": None}

    def validate_outdoor_seating(
        self,
        value: Text,
        dispatcher: CollectingDispatcher,
        tracker: Tracker,
        domain: Dict[Text, Any],
    ) -> Dict[Text, Any]:
        """Validate outdoor_seating value."""

        if isinstance(value, str):
            if "out" in value:
                # convert "out..." to True
                return {"outdoor_seating": True}
            elif "in" in value:
                # convert "in..." to False
                return {"outdoor_seating": False}
            else:
                dispatcher.utter_message(template="utter_wrong_outdoor_seating")
                # validation failed, set slot to None
                return {"outdoor_seating": None}

        else:
            # affirm/deny was picked up as T/F
            return {"outdoor_seating": value}

    def submit(
        self,
        dispatcher: CollectingDispatcher,
        tracker: Tracker,
        domain: Dict[Text, Any],
    ) -> List[Dict]:
        """Define what the form has to do
            after all required slots are filled"""

        # utter submit template
        dispatcher.utter_message(template="utter_submit")
        return []



import logging
from contextlib import contextmanager
from typing import Text, List

import pytest
from _pytest.logging import LogCaptureFixture
from _pytest.tmpdir import TempdirFactory
from sanic import Sanic

from rasa import server
from rasa.core import config
from rasa.core.agent import Agent, load_agent
from rasa.core.channels import channel
from rasa.core.channels.channel import RestInput
from rasa.core.policies import Policy
from rasa.core.policies.memoization import AugmentedMemoizationPolicy
from rasa.core.run import _create_app_without_api
from rasa.model import get_model
from rasa.train import train_async
from rasa.utils.common import TempDirectoryPath
from tests.core.conftest import (
    DEFAULT_DOMAIN_PATH_WITH_SLOTS,
    DEFAULT_NLU_DATA,
    DEFAULT_STACK_CONFIG,
    DEFAULT_STORIES_FILE,
    END_TO_END_STORY_FILE,
    MOODBOT_MODEL_PATH,
)

DEFAULT_CONFIG_PATH = "rasa/cli/default_config.yml"

# we reuse a bit of pytest's own testing machinery, this should eventually come
# from a separatedly installable pytest-cli plugin.
pytest_plugins = ["pytester"]


@pytest.fixture(autouse=True)
def set_log_level_debug(caplog: LogCaptureFixture) -> None:
    # Set the post-test log level to DEBUG for failing tests.  For all tests
    # (failing and successful), the live log level can be additionally set in
    # `setup.cfg`. It should be set to WARNING.
    caplog.set_level(logging.DEBUG)


@pytest.fixture
async def default_agent(tmpdir_factory: TempdirFactory) -> Agent:
    model_path = tmpdir_factory.mktemp("model").strpath

    agent = Agent(
        "data/test_domains/default.yml",
        policies=[AugmentedMemoizationPolicy(max_history=3)],
    )

    training_data = await agent.load_data(DEFAULT_STORIES_FILE)
    agent.train(training_data)
    agent.persist(model_path)
    return agent


@pytest.fixture(scope="session")
async def trained_moodbot_path() -> Text:
    return await train_async(
        domain="examples/moodbot/domain.yml",
        config="examples/moodbot/config.yml",
        training_files="examples/moodbot/data/",
        output_path=MOODBOT_MODEL_PATH,
    )


@pytest.fixture(scope="session")
async def unpacked_trained_moodbot_path(
    trained_moodbot_path: Text,
) -> TempDirectoryPath:
    return get_model(trained_moodbot_path)


@pytest.fixture
async def stack_agent(trained_rasa_model: Text) -> Agent:
    return await load_agent(model_path=trained_rasa_model)


@pytest.fixture
async def core_agent(trained_core_model: Text) -> Agent:
    return await load_agent(model_path=trained_core_model)


@pytest.fixture
async def nlu_agent(trained_nlu_model: Text) -> Agent:
    return await load_agent(model_path=trained_nlu_model)


@pytest.fixture(scope="session")
def default_domain_path() -> Text:
    return DEFAULT_DOMAIN_PATH_WITH_SLOTS


@pytest.fixture(scope="session")
def default_stories_file() -> Text:
    return DEFAULT_STORIES_FILE


@pytest.fixture(scope="session")
def default_stack_config() -> Text:
    return DEFAULT_STACK_CONFIG


@pytest.fixture(scope="session")
def default_nlu_data() -> Text:
    return DEFAULT_NLU_DATA


@pytest.fixture(scope="session")
def end_to_end_story_file() -> Text:
    return END_TO_END_STORY_FILE


@pytest.fixture(scope="session")
def default_config() -> List[Policy]:
    return config.load(DEFAULT_CONFIG_PATH)


@pytest.fixture(scope="session")
def trained_async(tmpdir_factory):
    async def _train(*args, output_path=None, **kwargs):
        if output_path is None:
            output_path = str(tmpdir_factory.mktemp("models"))

        return await train_async(*args, output_path=output_path, **kwargs)

    return _train


@pytest.fixture()
async def trained_rasa_model(
    trained_async,
    default_domain_path: Text,
    default_config: List[Policy],
    default_nlu_data: Text,
    default_stories_file: Text,
) -> Text:
    trained_stack_model_path = await trained_async(
        domain="data/test_domains/default.yml",
        config=DEFAULT_STACK_CONFIG,
        training_files=[default_nlu_data, default_stories_file],
    )

    return trained_stack_model_path


@pytest.fixture()
async def trained_core_model(
    trained_async,
    default_domain_path: Text,
    default_config: List[Policy],
    default_nlu_data: Text,
    default_stories_file: Text,
) -> Text:
    trained_core_model_path = await trained_async(
        domain=default_domain_path,
        config=DEFAULT_STACK_CONFIG,
        training_files=[default_stories_file],
    )

    return trained_core_model_path


@pytest.fixture()
async def trained_nlu_model(
    trained_async,
    default_domain_path: Text,
    default_config: List[Policy],
    default_nlu_data: Text,
    default_stories_file: Text,
) -> Text:
    trained_nlu_model_path = await trained_async(
        domain=default_domain_path,
        config=DEFAULT_STACK_CONFIG,
        training_files=[default_nlu_data],
    )

    return trained_nlu_model_path


@pytest.fixture
async def rasa_server(stack_agent: Agent) -> Sanic:
    app = server.create_app(agent=stack_agent)
    channel.register([RestInput()], app, "/webhooks/")
    return app


@pytest.fixture
async def rasa_core_server(core_agent: Agent) -> Sanic:
    app = server.create_app(agent=core_agent)
    channel.register([RestInput()], app, "/webhooks/")
    return app


@pytest.fixture
async def rasa_nlu_server(nlu_agent: Agent) -> Sanic:
    app = server.create_app(agent=nlu_agent)
    channel.register([RestInput()], app, "/webhooks/")
    return app


@pytest.fixture
async def rasa_server_secured(default_agent: Agent) -> Sanic:
    app = server.create_app(agent=default_agent, auth_token="rasa", jwt_secret="core")
    channel.register([RestInput()], app, "/webhooks/")
    return app


@pytest.fixture
async def rasa_server_without_api() -> Sanic:
    app = _create_app_without_api()
    channel.register([RestInput()], app, "/webhooks/")
    return app


def get_test_client(server):
    test_client = server.test_client
    test_client.port = None
    return test_client


@contextmanager
def assert_log_emitted(
    _caplog: LogCaptureFixture, logger_name: Text, log_level: int, text: Text = None
) -> None:
    """Context manager testing whether a logging message has been emitted.

    Provides a context in which an assertion is made about a logging message.
    Raises an `AssertionError` if the log isn't emitted as expected.

    Example usage:

    ```
    with assert_log_emitted(caplog, LOGGER_NAME, LOGGING_LEVEL, TEXT):
        <method supposed to emit TEXT at level LOGGING_LEVEL>
    ```

    Args:
        _caplog: `LogCaptureFixture` used to capture logs.
        logger_name: Name of the logger being examined.
        log_level: Log level to be tested.
        text: Logging message to be tested (optional). If left blank, assertion is made
            only about `log_level` and `logger_name`.

    Yields:
        `None`

    """

    yield

    record_tuples = _caplog.record_tuples

    if not any(
        (
            record[0] == logger_name
            and record[1] == log_level
            and (text in record[2] if text else True)
        )
        for record in record_tuples
    ):
        raise AssertionError(
            f"Did not detect expected logging output.\nExpected output is (logger "
            f"name, log level, text): ({logger_name}, {log_level}, {text})\n"
            f"Instead found records:\n{record_tuples}"
        )



import logging
import warnings
import os
import typing
from typing import Optional, Text

import rasa.utils.common as rasa_utils
from rasa.utils.endpoints import EndpointConfig

if typing.TYPE_CHECKING:
    from rasa.core.brokers.event_channel import EventChannel
    import pika

logger = logging.getLogger(__name__)


def from_endpoint_config(
    broker_config: Optional[EndpointConfig],
) -> Optional["EventChannel"]:
    """Instantiate an event channel based on its configuration."""

    if broker_config is None:
        return None
    elif broker_config.type == "pika" or broker_config.type is None:
        from rasa.core.brokers.pika import PikaProducer

        return PikaProducer.from_endpoint_config(broker_config)
    elif broker_config.type.lower() == "sql":
        from rasa.core.brokers.sql import SQLProducer

        return SQLProducer.from_endpoint_config(broker_config)
    elif broker_config.type == "file":
        from rasa.core.brokers.file_producer import FileProducer

        return FileProducer.from_endpoint_config(broker_config)
    elif broker_config.type == "kafka":
        from rasa.core.brokers.kafka import KafkaProducer

        return KafkaProducer.from_endpoint_config(broker_config)
    else:
        return load_event_channel_from_module_string(broker_config)


def load_event_channel_from_module_string(
    broker_config: EndpointConfig,
) -> Optional["EventChannel"]:
    """Instantiate an event channel based on its class name."""

    try:
        event_channel = rasa_utils.class_from_module_path(broker_config.type)
        return event_channel.from_endpoint_config(broker_config)
    except (AttributeError, ImportError) as e:
        warnings.warn(
            f"EventChannel type '{broker_config.type}' not found. "
            f"Not using any event channel. Error: {e}"
        )
        return None


def create_rabbitmq_ssl_options(
    rabbitmq_host: Optional[Text] = None,
) -> Optional["pika.SSLOptions"]:
    """Create RabbitMQ SSL options.

    Requires the following environment variables to be set:

        RABBITMQ_SSL_CLIENT_CERTIFICATE - path to the SSL client certificate (required)
        RABBITMQ_SSL_CLIENT_KEY - path to the SSL client key (required)
        RABBITMQ_SSL_CA_FILE - path to the SSL CA file for verification (optional)
        RABBITMQ_SSL_KEY_PASSWORD - SSL private key password (optional)

    Details on how to enable RabbitMQ TLS support can be found here:
    https://www.rabbitmq.com/ssl.html#enabling-tls

    Args:
        rabbitmq_host: RabbitMQ hostname

    Returns:
        Pika SSL context of type `pika.SSLOptions` if
        the RABBITMQ_SSL_CLIENT_CERTIFICATE and RABBITMQ_SSL_CLIENT_KEY
        environment variables are valid paths, else `None`.

    """

    client_certificate_path = os.environ.get("RABBITMQ_SSL_CLIENT_CERTIFICATE")
    client_key_path = os.environ.get("RABBITMQ_SSL_CLIENT_KEY")

    if client_certificate_path and client_key_path:
        import pika
        import rasa.server

        logger.debug(f"Configuring SSL context for RabbitMQ host '{rabbitmq_host}'.")

        ca_file_path = os.environ.get("RABBITMQ_SSL_CA_FILE")
        key_password = os.environ.get("RABBITMQ_SSL_KEY_PASSWORD")

        ssl_context = rasa.server.create_ssl_context(
            client_certificate_path, client_key_path, ca_file_path, key_password
        )
        return pika.SSLOptions(ssl_context, rabbitmq_host)
    else:
        return None



import asyncio
import logging
import time

logger = logging.getLogger(__name__)


def timeit(func):
    async def process(func, *args, **params):
        if asyncio.iscoroutinefunction(func):
            return await func(*args, **params)
        else:
            return func(*args, **params)

    async def helper(*args, **params):
        start = time.time()
        result = await process(func, *args, **params)
        run_time = round((time.time() - start) * 1000, 2)
        if run_time > 500:
            logger.warning(f"Function {func.__name__} finished in {run_time}ms")
        return result

    return helper



import os
import time
import tempfile
import uuid
from multiprocessing import Process, Manager
from typing import List, Text, Type
from contextlib import ExitStack

from aioresponses import aioresponses

import pytest
from freezegun import freeze_time
from mock import MagicMock

import rasa
import rasa.constants
from rasa.core import events, utils
from rasa.core.channels import CollectingOutputChannel, RestInput, SlackInput
from rasa.core.channels.slack import SlackBot
from rasa.core.events import Event, UserUttered, SlotSet, BotUttered
from rasa.core.trackers import DialogueStateTracker
from rasa.model import unpack_model
from rasa.utils.endpoints import EndpointConfig
from sanic import Sanic
from sanic.testing import SanicTestClient, PORT
from tests.nlu.utilities import ResponseTest
from tests.conftest import get_test_client


# a couple of event instances that we can use for testing
test_events = [
    Event.from_parameters(
        {
            "event": UserUttered.type_name,
            "text": "/goodbye",
            "parse_data": {
                "intent": {"confidence": 1.0, "name": "greet"},
                "entities": [],
            },
        }
    ),
    BotUttered("Welcome!", {"test": True}),
    SlotSet("cuisine", 34),
    SlotSet("cuisine", "34"),
    SlotSet("location", None),
    SlotSet("location", [34, "34", None]),
]


@pytest.fixture
def rasa_app_without_api(rasa_server_without_api: Sanic) -> SanicTestClient:
    return get_test_client(rasa_server_without_api)


@pytest.fixture
def rasa_app(rasa_server: Sanic) -> SanicTestClient:
    return get_test_client(rasa_server)


@pytest.fixture
def rasa_app_nlu(rasa_nlu_server: Sanic) -> SanicTestClient:
    return get_test_client(rasa_nlu_server)


@pytest.fixture
def rasa_app_core(rasa_core_server: Sanic) -> SanicTestClient:
    return get_test_client(rasa_core_server)


@pytest.fixture
def rasa_secured_app(rasa_server_secured: Sanic) -> SanicTestClient:
    return get_test_client(rasa_server_secured)


def test_root(rasa_app: SanicTestClient):
    _, response = rasa_app.get("/")
    assert response.status == 200
    assert response.text.startswith("Hello from Rasa:")


def test_root_without_enable_api(rasa_app_without_api: SanicTestClient):

    _, response = rasa_app_without_api.get("/")
    assert response.status == 200
    assert response.text.startswith("Hello from Rasa:")


def test_root_secured(rasa_secured_app: SanicTestClient):
    _, response = rasa_secured_app.get("/")
    assert response.status == 200
    assert response.text.startswith("Hello from Rasa:")


def test_version(rasa_app: SanicTestClient):
    _, response = rasa_app.get("/version")
    content = response.json
    assert response.status == 200
    assert content.get("version") == rasa.__version__
    assert (
        content.get("minimum_compatible_version")
        == rasa.constants.MINIMUM_COMPATIBLE_VERSION
    )


def test_status(rasa_app: SanicTestClient, trained_rasa_model: Text):
    _, response = rasa_app.get("/status")
    model_file = response.json["model_file"]
    assert response.status == 200
    assert "fingerprint" in response.json
    assert os.path.isfile(model_file)
    assert model_file == trained_rasa_model


def test_status_nlu_only(rasa_app_nlu: SanicTestClient, trained_nlu_model: Text):
    _, response = rasa_app_nlu.get("/status")
    model_file = response.json["model_file"]
    assert response.status == 200
    assert "fingerprint" in response.json
    assert "model_file" in response.json
    assert model_file == trained_nlu_model


def test_status_secured(rasa_secured_app: SanicTestClient):
    _, response = rasa_secured_app.get("/status")
    assert response.status == 401


def test_status_not_ready_agent(rasa_app: SanicTestClient):
    rasa_app.app.agent = None
    _, response = rasa_app.get("/status")
    assert response.status == 409


@pytest.fixture
def formbot_data():
    return dict(
        domain="examples/formbot/domain.yml",
        config="examples/formbot/config.yml",
        stories="examples/formbot/data/stories.md",
        nlu="examples/formbot/data/nlu.md",
    )


def test_train_status(rasa_server, rasa_app, formbot_data):
    with ExitStack() as stack:
        payload = {
            key: stack.enter_context(open(path)).read()
            for key, path in formbot_data.items()
        }

    def train(results):
        client1 = SanicTestClient(rasa_server, port=PORT + 1)
        _, train_resp = client1.post("/model/train", json=payload)
        results["train_response_code"] = train_resp.status

    # Run training process in the background
    manager = Manager()
    results = manager.dict()
    p1 = Process(target=train, args=(results,))
    p1.start()

    # Query the status endpoint a few times to ensure the test does
    # not fail prematurely due to mismatched timing of a single query.
    for i in range(10):
        time.sleep(1)
        _, status_resp = rasa_app.get("/status")
        assert status_resp.status == 200
        if status_resp.json["num_active_training_jobs"] == 1:
            break
    assert status_resp.json["num_active_training_jobs"] == 1

    p1.join()
    assert results["train_response_code"] == 200

    _, status_resp = rasa_app.get("/status")
    assert status_resp.status == 200
    assert status_resp.json["num_active_training_jobs"] == 0


@pytest.mark.parametrize(
    "response_test",
    [
        ResponseTest(
            "/model/parse",
            {
                "entities": [],
                "intent": {"confidence": 1.0, "name": "greet"},
                "text": "hello",
            },
            payload={"text": "hello"},
        ),
        ResponseTest(
            "/model/parse",
            {
                "entities": [],
                "intent": {"confidence": 1.0, "name": "greet"},
                "text": "hello",
            },
            payload={"text": "hello"},
        ),
        ResponseTest(
            "/model/parse",
            {
                "entities": [],
                "intent": {"confidence": 1.0, "name": "greet"},
                "text": "hello ńöñàśçií",
            },
            payload={"text": "hello ńöñàśçií"},
        ),
    ],
)
def test_parse(rasa_app, response_test):
    _, response = rasa_app.post(response_test.endpoint, json=response_test.payload)
    rjs = response.json
    assert response.status == 200
    assert all(prop in rjs for prop in ["entities", "intent", "text"])
    assert rjs["entities"] == response_test.expected_response["entities"]
    assert rjs["text"] == response_test.expected_response["text"]
    assert rjs["intent"] == response_test.expected_response["intent"]


@pytest.mark.parametrize(
    "response_test",
    [
        ResponseTest(
            "/model/parse?emulation_mode=wit",
            {
                "entities": [],
                "intent": {"confidence": 1.0, "name": "greet"},
                "text": "hello",
            },
            payload={"text": "hello"},
        ),
        ResponseTest(
            "/model/parse?emulation_mode=dialogflow",
            {
                "entities": [],
                "intent": {"confidence": 1.0, "name": "greet"},
                "text": "hello",
            },
            payload={"text": "hello"},
        ),
        ResponseTest(
            "/model/parse?emulation_mode=luis",
            {
                "entities": [],
                "intent": {"confidence": 1.0, "name": "greet"},
                "text": "hello ńöñàśçií",
            },
            payload={"text": "hello ńöñàśçií"},
        ),
    ],
)
def test_parse_with_different_emulation_mode(rasa_app, response_test):
    _, response = rasa_app.post(response_test.endpoint, json=response_test.payload)
    assert response.status == 200


def test_parse_without_nlu_model(rasa_app_core: SanicTestClient):
    _, response = rasa_app_core.post("/model/parse", json={"text": "hello"})
    assert response.status == 200

    rjs = response.json
    assert all(prop in rjs for prop in ["entities", "intent", "text"])


def test_parse_on_invalid_emulation_mode(rasa_app_nlu: SanicTestClient):
    _, response = rasa_app_nlu.post(
        "/model/parse?emulation_mode=ANYTHING", json={"text": "hello"}
    )
    assert response.status == 400


def test_train_stack_success(
    rasa_app,
    default_domain_path,
    default_stories_file,
    default_stack_config,
    default_nlu_data,
):
    with ExitStack() as stack:
        domain_file = stack.enter_context(open(default_domain_path, encoding='utf8'))
        config_file = stack.enter_context(open(default_stack_config))
        stories_file = stack.enter_context(open(default_stories_file))
        nlu_file = stack.enter_context(open(default_nlu_data))

        payload = dict(
            domain=domain_file.read(),
            config=config_file.read(),
            stories=stories_file.read(),
            nlu=nlu_file.read(),
        )

    _, response = rasa_app.post("/model/train", json=payload)
    assert response.status == 200

    assert response.headers["filename"] is not None

    # save model to temporary file
    tempdir = tempfile.mkdtemp()
    model_path = os.path.join(tempdir, "model.tar.gz")
    with open(model_path, "wb") as f:
        f.write(response.body)

    # unpack model and ensure fingerprint is present
    model_path = unpack_model(model_path)
    assert os.path.exists(os.path.join(model_path, "fingerprint.json"))


def test_train_nlu_success(
    rasa_app, default_stack_config, default_nlu_data, default_domain_path
):
    with ExitStack() as stack:
        domain_file = stack.enter_context(open(default_domain_path, encoding='utf8'))
        config_file = stack.enter_context(open(default_stack_config))
        nlu_file = stack.enter_context(open(default_nlu_data))

        payload = dict(
            domain=domain_file.read(), config=config_file.read(), nlu=nlu_file.read()
        )

    _, response = rasa_app.post("/model/train", json=payload)
    assert response.status == 200

    # save model to temporary file
    tempdir = tempfile.mkdtemp()
    model_path = os.path.join(tempdir, "model.tar.gz")
    with open(model_path, "wb") as f:
        f.write(response.body)

    # unpack model and ensure fingerprint is present
    model_path = unpack_model(model_path)
    assert os.path.exists(os.path.join(model_path, "fingerprint.json"))


def test_train_core_success(
    rasa_app, default_stack_config, default_stories_file, default_domain_path
):
    with ExitStack() as stack:
        domain_file = stack.enter_context(open(default_domain_path, encoding='utf8'))
        config_file = stack.enter_context(open(default_stack_config))
        core_file = stack.enter_context(open(default_stories_file))

        payload = dict(
            domain=domain_file.read(),
            config=config_file.read(),
            stories=core_file.read(),
        )

    _, response = rasa_app.post("/model/train", json=payload)
    assert response.status == 200

    # save model to temporary file
    tempdir = tempfile.mkdtemp()
    model_path = os.path.join(tempdir, "model.tar.gz")
    with open(model_path, "wb") as f:
        f.write(response.body)

    # unpack model and ensure fingerprint is present
    model_path = unpack_model(model_path)
    assert os.path.exists(os.path.join(model_path, "fingerprint.json"))


def test_train_missing_config(rasa_app: SanicTestClient):
    payload = dict(domain="domain data", config=None)

    _, response = rasa_app.post("/model/train", json=payload)
    assert response.status == 400


def test_train_missing_training_data(rasa_app: SanicTestClient):
    payload = dict(domain="domain data", config="config data")

    _, response = rasa_app.post("/model/train", json=payload)
    assert response.status == 400


def test_train_internal_error(rasa_app: SanicTestClient):
    payload = dict(domain="domain data", config="config data", nlu="nlu data")

    _, response = rasa_app.post("/model/train", json=payload)
    assert response.status == 500


def test_evaluate_stories(rasa_app, default_stories_file):
    with open(default_stories_file, "r") as f:
        stories = f.read()

    _, response = rasa_app.post("/model/test/stories", data=stories)

    assert response.status == 200

    js = response.json
    assert set(js.keys()) == {
        "report",
        "precision",
        "f1",
        "accuracy",
        "actions",
        "in_training_data_fraction",
        "is_end_to_end_evaluation",
    }
    assert not js["is_end_to_end_evaluation"]
    assert set(js["actions"][0].keys()) == {
        "action",
        "predicted",
        "confidence",
        "policy",
    }


def test_evaluate_stories_not_ready_agent(
    rasa_app_nlu: SanicTestClient, default_stories_file
):
    with open(default_stories_file, "r") as f:
        stories = f.read()

    _, response = rasa_app_nlu.post("/model/test/stories", data=stories)

    assert response.status == 409


def test_evaluate_stories_end_to_end(rasa_app, end_to_end_story_file):
    with open(end_to_end_story_file, "r") as f:
        stories = f.read()

    _, response = rasa_app.post("/model/test/stories?e2e=true", data=stories)

    assert response.status == 200
    js = response.json
    assert set(js.keys()) == {
        "report",
        "precision",
        "f1",
        "accuracy",
        "actions",
        "in_training_data_fraction",
        "is_end_to_end_evaluation",
    }
    assert js["is_end_to_end_evaluation"]
    assert set(js["actions"][0].keys()) == {
        "action",
        "predicted",
        "confidence",
        "policy",
    }


def test_evaluate_intent(rasa_app, default_nlu_data):
    with open(default_nlu_data, "r") as f:
        nlu_data = f.read()

    _, response = rasa_app.post("/model/test/intents", data=nlu_data)

    assert response.status == 200
    assert set(response.json.keys()) == {
        "intent_evaluation",
        "entity_evaluation",
        "response_selection_evaluation",
    }


def test_evaluate_intent_on_just_nlu_model(
    rasa_app_nlu: SanicTestClient, default_nlu_data
):
    with open(default_nlu_data, "r") as f:
        nlu_data = f.read()

    _, response = rasa_app_nlu.post("/model/test/intents", data=nlu_data)

    assert response.status == 200
    assert set(response.json.keys()) == {
        "intent_evaluation",
        "entity_evaluation",
        "response_selection_evaluation",
    }


def test_evaluate_intent_with_query_param(
    rasa_app, trained_nlu_model, default_nlu_data
):
    _, response = rasa_app.get("/status")
    previous_model_file = response.json["model_file"]

    with open(default_nlu_data, "r") as f:
        nlu_data = f.read()

    _, response = rasa_app.post(
        f"/model/test/intents?model={trained_nlu_model}", data=nlu_data
    )

    assert response.status == 200
    assert set(response.json.keys()) == {
        "intent_evaluation",
        "entity_evaluation",
        "response_selection_evaluation",
    }

    _, response = rasa_app.get("/status")
    assert previous_model_file == response.json["model_file"]


def test_predict(rasa_app: SanicTestClient):
    data = {
        "Events": {
            "value": [
                {"event": "action", "name": "action_listen"},
                {
                    "event": "user",
                    "text": "hello",
                    "parse_data": {
                        "entities": [],
                        "intent": {"confidence": 0.57, "name": "greet"},
                        "text": "hello",
                    },
                },
            ]
        }
    }
    _, response = rasa_app.post(
        "/model/predict", json=data, headers={"Content-Type": "application/json"}
    )
    content = response.json
    assert response.status == 200
    assert "scores" in content
    assert "tracker" in content
    assert "policy" in content


@freeze_time("2018-01-01")
def test_requesting_non_existent_tracker(rasa_app: SanicTestClient):
    _, response = rasa_app.get("/conversations/madeupid/tracker")
    content = response.json
    assert response.status == 200
    assert content["paused"] is False
    assert content["slots"] == {"location": None, "cuisine": None}
    assert content["sender_id"] == "madeupid"
    assert content["events"] == [
        {
            "event": "action",
            "name": "action_listen",
            "policy": None,
            "confidence": None,
            "timestamp": 1514764800,
        }
    ]
    assert content["latest_message"] == {
        "text": None,
        "intent": {},
        "entities": [],
        "message_id": None,
        "asrConfidence": 1,
        "missed_intent": None,
        "recordingId": None,
        "metadata": {},
    }


def test_check_tracker_existence(rasa_app: SanicTestClient):
    _, response = rasa_app.get("/conversations/madeupid/tracker/exists")
    content = response.json
    assert response.status == 200
    assert content["sender_id"] == "madeupid"
    assert content["exists"] == False

    rasa_app.get("/conversations/madeupid/tracker")
    _, response = rasa_app.get("/conversations/madeupid/tracker/exists")
    content = response.json
    assert response.status == 200
    assert content["sender_id"] == "madeupid"
    assert content["exists"] == True


@pytest.mark.parametrize("event", test_events)
def test_pushing_event(rasa_app, event):
    cid = str(uuid.uuid1())
    conversation = f"/conversations/{cid}"

    _, response = rasa_app.post(
        f"{conversation}/tracker/events",
        json=event.as_dict(),
        headers={"Content-Type": "application/json"},
    )
    assert response.json is not None
    assert response.status == 200

    _, tracker_response = rasa_app.get(f"/conversations/{cid}/tracker")
    tracker = tracker_response.json
    assert tracker is not None
    assert len(tracker.get("events")) == 2

    evt = tracker.get("events")[1]
    assert Event.from_parameters(evt) == event


def test_push_multiple_events(rasa_app: SanicTestClient):
    cid = str(uuid.uuid1())
    conversation = f"/conversations/{cid}"

    events = [e.as_dict() for e in test_events]
    _, response = rasa_app.post(
        f"{conversation}/tracker/events",
        json=events,
        headers={"Content-Type": "application/json"},
    )
    assert response.json is not None
    assert response.status == 200

    _, tracker_response = rasa_app.get(f"/conversations/{cid}/tracker")
    tracker = tracker_response.json
    assert tracker is not None

    # there is also an `ACTION_LISTEN` event at the start
    assert len(tracker.get("events")) == len(test_events) + 1
    assert tracker.get("events")[1:] == events


def test_put_tracker(rasa_app: SanicTestClient):
    data = [event.as_dict() for event in test_events]
    _, response = rasa_app.put(
        "/conversations/pushtracker/tracker/events",
        json=data,
        headers={"Content-Type": "application/json"},
    )
    content = response.json
    assert response.status == 200
    assert len(content["events"]) == len(test_events)
    assert content["sender_id"] == "pushtracker"

    _, tracker_response = rasa_app.get("/conversations/pushtracker/tracker")
    tracker = tracker_response.json
    assert tracker is not None
    evts = tracker.get("events")
    assert events.deserialise_events(evts) == test_events


def test_sorted_predict(rasa_app: SanicTestClient):
    _create_tracker_for_sender(rasa_app, "sortedpredict")

    _, response = rasa_app.post("/conversations/sortedpredict/predict")
    scores = response.json["scores"]
    sorted_scores = sorted(scores, key=lambda k: (-k["score"], k["action"]))
    assert scores == sorted_scores


def _create_tracker_for_sender(app: SanicTestClient, sender_id: Text) -> None:
    data = [event.as_dict() for event in test_events[:3]]
    _, response = app.put(
        f"/conversations/{sender_id}/tracker/events",
        json=data,
        headers={"Content-Type": "application/json"},
    )

    assert response.status == 200


def test_get_tracker_with_jwt(rasa_secured_app):
    # token generated with secret "core" and algorithm HS256
    # on https://jwt.io/

    # {"user": {"username": "testadmin", "role": "admin"}}
    jwt_header = {
        "Authorization": "Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9."
        "eyJ1c2VyIjp7InVzZXJuYW1lIjoidGVzdGFkbWluIiwic"
        "m9sZSI6ImFkbWluIn19.NAQr0kbtSrY7d28XTqRzawq2u"
        "QRre7IWTuIDrCn5AIw"
    }
    _, response = rasa_secured_app.get(
        "/conversations/testadmin/tracker", headers=jwt_header
    )
    assert response.status == 200

    _, response = rasa_secured_app.get(
        "/conversations/testuser/tracker", headers=jwt_header
    )
    assert response.status == 200

    # {"user": {"username": "testuser", "role": "user"}}
    jwt_header = {
        "Authorization": "Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9."
        "eyJ1c2VyIjp7InVzZXJuYW1lIjoidGVzdHVzZXIiLCJyb"
        "2xlIjoidXNlciJ9fQ.JnMTLYd56qut2w9h7hRQlDm1n3l"
        "HJHOxxC_w7TtwCrs"
    }
    _, response = rasa_secured_app.get(
        "/conversations/testadmin/tracker", headers=jwt_header
    )
    assert response.status == 403

    _, response = rasa_secured_app.get(
        "/conversations/testuser/tracker", headers=jwt_header
    )
    assert response.status == 200


def test_list_routes(default_agent):
    from rasa import server

    app = server.create_app(default_agent, auth_token=None)

    routes = utils.list_routes(app)
    assert {
        "hello",
        "version",
        "status",
        "retrieve_tracker",
        "append_events",
        "replace_events",
        "retrieve_story",
        "execute_action",
        "predict",
        "add_message",
        "train",
        "evaluate_stories",
        "evaluate_intents",
        "tracker_predict",
        "parse",
        "load_model",
        "unload_model",
        "parseList",
        "ready",
        "live",
        "process_response",
        "get_domain",
    }.issubset(set(routes.keys()))


def test_unload_model_error(rasa_app: SanicTestClient):
    _, response = rasa_app.get("/status")
    assert response.status == 200
    assert "model_file" in response.json and response.json["model_file"] is not None

    _, response = rasa_app.delete("/model")
    assert response.status == 204


def test_get_domain(rasa_app: SanicTestClient):
    _, response = rasa_app.get("/domain", headers={"accept": "application/json"})

    content = response.json

    assert response.status == 200
    assert "config" in content
    assert "intents" in content
    assert "entities" in content
    assert "slots" in content
    assert "templates" in content
    assert "actions" in content


def test_get_domain_invalid_accept_header(rasa_app: SanicTestClient):
    _, response = rasa_app.get("/domain")

    assert response.status == 406


def test_load_model(rasa_app: SanicTestClient, trained_core_model):
    _, response = rasa_app.get("/status")

    assert response.status == 200
    assert "fingerprint" in response.json

    old_fingerprint = response.json["fingerprint"]

    data = {"model_file": trained_core_model}
    _, response = rasa_app.put("/model", json=data)

    assert response.status == 204

    _, response = rasa_app.get("/status")

    assert response.status == 200
    assert "fingerprint" in response.json

    assert old_fingerprint != response.json["fingerprint"]


def test_load_model_from_model_server(rasa_app: SanicTestClient, trained_core_model):
    _, response = rasa_app.get("/status")

    assert response.status == 200
    assert "fingerprint" in response.json

    old_fingerprint = response.json["fingerprint"]

    endpoint = EndpointConfig("https://example.com/model/trained_core_model")
    with open(trained_core_model, "rb") as f:
        with aioresponses(passthrough=["http://127.0.0.1"]) as mocked:
            headers = {}
            fs = os.fstat(f.fileno())
            headers["Content-Length"] = str(fs[6])
            mocked.get(
                "https://example.com/model/trained_core_model",
                content_type="application/x-tar",
                body=f.read(),
            )
            data = {"model_server": {"url": endpoint.url}}
            _, response = rasa_app.put("/model", json=data)

            assert response.status == 204

            _, response = rasa_app.get("/status")

            assert response.status == 200
            assert "fingerprint" in response.json

            assert old_fingerprint != response.json["fingerprint"]

    import rasa.core.jobs

    rasa.core.jobs.__scheduler = None


def test_load_model_invalid_request_body(rasa_app: SanicTestClient):
    _, response = rasa_app.put("/model")

    assert response.status == 400


def test_load_model_invalid_configuration(rasa_app: SanicTestClient):
    data = {"model_file": "some-random-path"}
    _, response = rasa_app.put("/model", json=data)

    assert response.status == 400


def test_execute(rasa_app: SanicTestClient):
    _create_tracker_for_sender(rasa_app, "test_execute")

    data = {"name": "utter_greet"}
    _, response = rasa_app.post("/conversations/test_execute/execute", json=data)

    assert response.status == 200

    parsed_content = response.json
    assert parsed_content["tracker"]
    assert parsed_content["messages"]


def test_execute_with_missing_action_name(rasa_app: SanicTestClient):
    test_sender = "test_execute_with_missing_action_name"
    _create_tracker_for_sender(rasa_app, test_sender)

    data = {"wrong-key": "utter_greet"}
    _, response = rasa_app.post(f"/conversations/{test_sender}/execute", json=data)

    assert response.status == 400


def test_execute_with_not_existing_action(rasa_app: SanicTestClient):
    test_sender = "test_execute_with_not_existing_action"
    _create_tracker_for_sender(rasa_app, test_sender)

    data = {"name": "ka[pa[opi[opj[oj[oija"}
    _, response = rasa_app.post(f"/conversations/{test_sender}/execute", json=data)

    assert response.status == 500


@pytest.mark.parametrize(
    "input_channels, output_channel_to_use, expected_channel",
    [
        (None, "slack", CollectingOutputChannel),
        ([], None, CollectingOutputChannel),
        ([RestInput()], "slack", CollectingOutputChannel),
        ([RestInput()], "rest", CollectingOutputChannel),
        ([RestInput(), SlackInput("test")], "slack", SlackBot),
    ],
)
def test_get_output_channel(
    input_channels: List[Text], output_channel_to_use, expected_channel: Type
):
    request = MagicMock()
    app = MagicMock()
    app.input_channels = input_channels
    request.app = app
    request.args = {"output_channel": output_channel_to_use}

    actual = rasa.server._get_output_channel(request, None)

    assert isinstance(actual, expected_channel)


@pytest.mark.parametrize(
    "input_channels, expected_channel",
    [
        ([], CollectingOutputChannel),
        ([RestInput()], CollectingOutputChannel),
        ([RestInput(), SlackInput("test")], SlackBot),
    ],
)
def test_get_latest_output_channel(input_channels: List[Text], expected_channel: Type):
    request = MagicMock()
    app = MagicMock()
    app.input_channels = input_channels
    request.app = app
    request.args = {"output_channel": "latest"}

    tracker = DialogueStateTracker.from_events(
        "default", [UserUttered("text", input_channel="slack")]
    )

    actual = rasa.server._get_output_channel(request, tracker)

    assert isinstance(actual, expected_channel)


def test_app_when_app_has_no_input_channels():
    request = MagicMock()

    class NoInputChannels:
        pass

    request.app = NoInputChannels()

    actual = rasa.server._get_output_channel(
        request, DialogueStateTracker.from_events("default", [])
    )
    assert isinstance(actual, CollectingOutputChannel)



import json
import logging
import sys
import uuid
from collections import deque, defaultdict
from typing import List, Text, Dict, Optional, Tuple, Any, Set, ValuesView

from rasa.core import utils
from rasa.core.actions.action import ACTION_LISTEN_NAME
from rasa.core.conversation import Dialogue
from rasa.core.domain import Domain
from rasa.core.events import (
    UserUttered,
    ActionExecuted,
    Form,
    FormValidation,
    SlotSet,
    Event,
    ActionExecutionRejected,
)

logger = logging.getLogger(__name__)

# Checkpoint id used to identify story starting blocks
STORY_START = "STORY_START"

# Checkpoint id used to identify story end blocks
STORY_END = None

# need abbreviations otherwise they are not visualized well
GENERATED_CHECKPOINT_PREFIX = "GENR_"
CHECKPOINT_CYCLE_PREFIX = "CYCL_"

GENERATED_HASH_LENGTH = 5

FORM_PREFIX = "form: "
# prefix for storystep ID to get reproducible sorting results
# will get increased with each new instance
STEP_COUNT = 1


class StoryStringHelper:
    """A helper class to mark story steps that are inside a form with `form: `
    """

    def __init__(
        self,
        active_form=None,
        form_validation=True,
        form_rejected=False,
        form_prefix_string="",
        no_form_prefix_string="",
    ):
        # track active form
        self.active_form = active_form
        # track whether a from should be validated
        self.form_validation = form_validation
        # track whether a from was rejected
        self.form_rejected = form_rejected
        # save story strings with form prefix for later
        self.form_prefix_string = form_prefix_string
        # save story strings without form prefix for later
        self.no_form_prefix_string = no_form_prefix_string


class Checkpoint:
    def __init__(
        self, name: Optional[Text], conditions: Optional[Dict[Text, Any]] = None
    ) -> None:

        self.name = name
        self.conditions = conditions if conditions else {}

    def as_story_string(self):
        dumped_conds = json.dumps(self.conditions) if self.conditions else ""
        return f"{self.name}{dumped_conds}"

    def filter_trackers(self, trackers):
        """Filters out all trackers that do not satisfy the conditions."""

        if not self.conditions:
            return trackers

        for slot_name, slot_value in self.conditions.items():
            trackers = [t for t in trackers if t.get_slot(slot_name) == slot_value]
        return trackers

    def __repr__(self):
        return "Checkpoint(name={!r}, conditions={})".format(
            self.name, json.dumps(self.conditions)
        )


class StoryStep:
    """A StoryStep is a section of a story block between two checkpoints.

    NOTE: Checkpoints are not only limited to those manually written
    in the story file, but are also implicitly created at points where
    multiple intents are separated in one line by chaining them with "OR"s.
    """

    def __init__(
        self,
        block_name: Optional[Text] = None,
        start_checkpoints: Optional[List[Checkpoint]] = None,
        end_checkpoints: Optional[List[Checkpoint]] = None,
        events: Optional[List[Event]] = None,
    ) -> None:

        self.end_checkpoints = end_checkpoints if end_checkpoints else []
        self.start_checkpoints = start_checkpoints if start_checkpoints else []
        self.events = events if events else []
        self.block_name = block_name
        # put a counter prefix to uuid to get reproducible sorting results
        global STEP_COUNT
        self.id = "{}_{}".format(STEP_COUNT, uuid.uuid4().hex)
        STEP_COUNT += 1

        self.story_string_helper = StoryStringHelper()

    def create_copy(self, use_new_id):
        copied = StoryStep(
            self.block_name,
            self.start_checkpoints,
            self.end_checkpoints,
            self.events[:],
        )
        if not use_new_id:
            copied.id = self.id
        return copied

    def add_user_message(self, user_message):
        self.add_event(user_message)

    def add_event(self, event):
        self.events.append(event)

    @staticmethod
    def _checkpoint_string(story_step_element: UserUttered) -> Text:
        return "> {}\n".format(story_step_element.as_story_string())

    @staticmethod
    def _user_string(
        story_step_element: UserUttered, e2e: bool, prefix: Text = ""
    ) -> Text:
        return "* {}{}\n".format(prefix, story_step_element.as_story_string(e2e))

    def _store_user_strings(
        self, story_step_element: UserUttered, e2e: bool, prefix: Text = ""
    ) -> None:
        self.story_string_helper.no_form_prefix_string += self._user_string(
            story_step_element, e2e
        )
        self.story_string_helper.form_prefix_string += self._user_string(
            story_step_element, e2e, prefix
        )

    @staticmethod
    def _bot_string(story_step_element, prefix=""):
        return "    - {}{}\n".format(prefix, story_step_element.as_story_string())

    def _store_bot_strings(self, story_step_element, prefix=""):
        self.story_string_helper.no_form_prefix_string += self._bot_string(
            story_step_element
        )
        self.story_string_helper.form_prefix_string += self._bot_string(
            story_step_element, prefix
        )

    def _reset_stored_strings(self):
        self.story_string_helper.form_prefix_string = ""
        self.story_string_helper.no_form_prefix_string = ""

    def as_story_string(self, flat: bool = False, e2e: bool = False) -> Text:
        # if the result should be flattened, we
        # will exclude the caption and any checkpoints.

        for s in self.start_checkpoints:
            if s.name == STORY_START:
                # first story step in the story, so reset helper
                self.story_string_helper = StoryStringHelper()

        if flat:
            result = ""
        else:
            result = f"\n## {self.block_name}\n"
            for s in self.start_checkpoints:
                if s.name != STORY_START:
                    result += self._checkpoint_string(s)

        for s in self.events:
            if isinstance(s, UserUttered):
                if self.story_string_helper.active_form is None:
                    result += self._user_string(s, e2e)
                else:
                    # form is active
                    # it is not known whether the form will be
                    # successfully executed, so store this
                    # story string for later
                    self._store_user_strings(s, e2e, FORM_PREFIX)

            elif isinstance(s, Form):
                # form got either activated or deactivated
                self.story_string_helper.active_form = s.name

                if self.story_string_helper.active_form is None:
                    # form deactivated, so form succeeded,
                    # so add story string with form prefix
                    result += self.story_string_helper.form_prefix_string
                    # remove all stored story strings
                    self._reset_stored_strings()

                result += self._bot_string(s)

            elif isinstance(s, FormValidation):
                self.story_string_helper.form_validation = s.validate

            elif isinstance(s, ActionExecutionRejected):
                if s.action_name == self.story_string_helper.active_form:
                    # form rejected
                    self.story_string_helper.form_rejected = True

            elif isinstance(s, ActionExecuted):
                if self._is_action_listen(s):
                    pass
                elif self.story_string_helper.active_form is None:
                    result += self._bot_string(s)
                else:
                    # form is active
                    if self.story_string_helper.form_rejected:
                        if (
                            self.story_string_helper.form_prefix_string
                            and self.story_string_helper.form_validation
                            and s.action_name == self.story_string_helper.active_form
                        ):
                            # if there is something in `form_prefix_string`,
                            # add action_listen before it,
                            # because this form user input will be ignored by core
                            # and therefore action_listen will not be automatically
                            # added during reading the stories
                            result += self._bot_string(
                                ActionExecuted(ACTION_LISTEN_NAME)
                            )
                            result += self.story_string_helper.form_prefix_string
                        elif self.story_string_helper.no_form_prefix_string:
                            result += self.story_string_helper.no_form_prefix_string
                        # form rejected, add story string without form prefix
                        result += self._bot_string(s)
                    else:
                        # form succeeded, so add story string with form prefix
                        result += self.story_string_helper.form_prefix_string
                        result += self._bot_string(s, FORM_PREFIX)

                    # remove all stored story strings
                    self._reset_stored_strings()

                    if s.action_name == self.story_string_helper.active_form:
                        # form was successfully executed
                        self.story_string_helper.form_rejected = False

                self.story_string_helper.form_validation = True

            elif isinstance(s, SlotSet):
                if self.story_string_helper.active_form is None:
                    result += self._bot_string(s)
                else:
                    # form is active
                    # it is not known whether the form will be
                    # successfully executed, so store this
                    # story string for later
                    # slots should be always printed without prefix
                    self._store_bot_strings(s)

            elif isinstance(s, Event):
                converted = s.as_story_string()
                if converted:
                    if self.story_string_helper.active_form is None:
                        result += self._bot_string(s)
                    else:
                        # form is active
                        # it is not known whether the form will be
                        # successfully executed, so store this
                        # story string for later
                        self._store_bot_strings(s, FORM_PREFIX)

            else:
                raise Exception(f"Unexpected element in story step: {s}")

        if (
            not self.end_checkpoints
            and self.story_string_helper.active_form is not None
        ):
            # there are no end checkpoints
            # form is active
            # add story string with form prefix
            result += self.story_string_helper.form_prefix_string
            # remove all stored story strings
            self._reset_stored_strings()

        if not flat:
            for e in self.end_checkpoints:
                result += "> {}\n".format(e.as_story_string())
        return result

    @staticmethod
    def _is_action_listen(event):
        # this is not an `isinstance` because
        # we don't want to allow subclasses here
        return type(event) == ActionExecuted and event.action_name == ACTION_LISTEN_NAME

    def _add_action_listen(self, events):
        if not events or not self._is_action_listen(events[-1]):
            # do not add second action_listen
            events.append(ActionExecuted(ACTION_LISTEN_NAME))

    def explicit_events(
        self, domain: Domain, should_append_final_listen: bool = True
    ) -> List[Event]:
        """Returns events contained in the story step
            including implicit events.

        Not all events are always listed in the story dsl. This
        includes listen actions as well as implicitly
        set slots. This functions makes these events explicit and
        returns them with the rest of the steps events."""

        events = []

        for e in self.events:
            if isinstance(e, UserUttered):
                self._add_action_listen(events)
                events.append(e)
                events.extend(domain.slots_for_entities(e.entities))
            else:
                events.append(e)

        if not self.end_checkpoints and should_append_final_listen:
            self._add_action_listen(events)

        return events

    def __repr__(self):
        return (
            "StoryStep("
            "block_name={!r}, "
            "start_checkpoints={!r}, "
            "end_checkpoints={!r}, "
            "events={!r})".format(
                self.block_name,
                self.start_checkpoints,
                self.end_checkpoints,
                self.events,
            )
        )


class Story:
    def __init__(
        self, story_steps: List[StoryStep] = None, story_name: Optional[Text] = None
    ) -> None:
        self.story_steps = story_steps if story_steps else []
        self.story_name = story_name

    @staticmethod
    def from_events(events, story_name=None):
        """Create a story from a list of events."""

        story_step = StoryStep()
        for event in events:
            story_step.add_event(event)
        return Story([story_step], story_name)

    def as_dialogue(self, sender_id, domain):
        events = []
        for step in self.story_steps:
            events.extend(
                step.explicit_events(domain, should_append_final_listen=False)
            )

        events.append(ActionExecuted(ACTION_LISTEN_NAME))
        return Dialogue(sender_id, events)

    def as_story_string(self, flat: bool = False, e2e: bool = False) -> Text:
        story_content = ""

        # initialize helper for first story step
        story_string_helper = StoryStringHelper()

        for step in self.story_steps:
            # use helper from previous story step
            step.story_string_helper = story_string_helper
            # create string for current story step
            story_content += step.as_story_string(flat, e2e)
            # override helper for next story step
            story_string_helper = step.story_string_helper

        if flat:
            if self.story_name:
                name = self.story_name
            else:
                name = "Generated Story {}".format(hash(story_content))
            return f"## {name}\n{story_content}"
        else:
            return story_content

    def dump_to_file(
        self, filename: Text, flat: bool = False, e2e: bool = False
    ) -> None:
        from rasa.utils import io

        io.write_text_file(self.as_story_string(flat, e2e), filename, append=True)


class StoryGraph:
    """Graph of the story-steps pooled from all stories in the training data."""

    def __init__(
        self,
        story_steps: List[StoryStep],
        story_end_checkpoints: Optional[Dict[Text, Text]] = None,
    ) -> None:
        self.story_steps = story_steps
        self.step_lookup = {s.id: s for s in self.story_steps}
        ordered_ids, cyclic_edges = StoryGraph.order_steps(story_steps)
        self.ordered_ids = ordered_ids
        self.cyclic_edge_ids = cyclic_edges
        if story_end_checkpoints:
            self.story_end_checkpoints = story_end_checkpoints
        else:
            self.story_end_checkpoints = {}

    def __hash__(self) -> int:
        self_as_string = self.as_story_string()
        text_hash = utils.get_text_hash(self_as_string)

        return int(text_hash, 16)

    def ordered_steps(self) -> List[StoryStep]:
        """Returns the story steps ordered by topological order of the DAG."""

        return [self.get(step_id) for step_id in self.ordered_ids]

    def cyclic_edges(self) -> List[Tuple[Optional[StoryStep], Optional[StoryStep]]]:
        """Returns the story steps ordered by topological order of the DAG."""

        return [
            (self.get(source), self.get(target))
            for source, target in self.cyclic_edge_ids
        ]

    def merge(self, other: Optional["StoryGraph"]) -> "StoryGraph":
        if not other:
            return self

        steps = self.story_steps.copy() + other.story_steps
        story_end_checkpoints = self.story_end_checkpoints.copy().update(
            other.story_end_checkpoints
        )
        return StoryGraph(steps, story_end_checkpoints)

    @staticmethod
    def overlapping_checkpoint_names(
        cps: List[Checkpoint], other_cps: List[Checkpoint]
    ) -> Set[Text]:
        """Find overlapping checkpoints names"""

        return {cp.name for cp in cps} & {cp.name for cp in other_cps}

    def with_cycles_removed(self) -> "StoryGraph":
        """Create a graph with the cyclic edges removed from this graph."""

        story_end_checkpoints = self.story_end_checkpoints.copy()
        cyclic_edge_ids = self.cyclic_edge_ids
        # we need to remove the start steps and replace them with steps ending
        # in a special end checkpoint

        story_steps = {s.id: s for s in self.story_steps}

        # collect all overlapping checkpoints
        # we will remove unused start ones
        all_overlapping_cps = set()

        if self.cyclic_edge_ids:
            # we are going to do this in a recursive way. we are going to
            # remove one cycle and then we are going to
            # let the cycle detection run again
            # this is not inherently necessary so if this becomes a performance
            # issue, we can change it. It is actually enough to run the cycle
            # detection only once and then remove one cycle after another, but
            # since removing the cycle is done by adding / removing edges and
            #  nodes
            # the logic is a lot easier if we only need to make sure the
            # change is consistent if we only change one compared to
            # changing all of them.

            for s, e in cyclic_edge_ids:
                cid = utils.generate_id(max_chars=GENERATED_HASH_LENGTH)
                prefix = GENERATED_CHECKPOINT_PREFIX + CHECKPOINT_CYCLE_PREFIX
                # need abbreviations otherwise they are not visualized well
                sink_cp_name = prefix + "SINK_" + cid
                connector_cp_name = prefix + "CONN_" + cid
                source_cp_name = prefix + "SRC_" + cid
                story_end_checkpoints[sink_cp_name] = source_cp_name

                overlapping_cps = self.overlapping_checkpoint_names(
                    story_steps[s].end_checkpoints, story_steps[e].start_checkpoints
                )

                all_overlapping_cps.update(overlapping_cps)

                # change end checkpoints of starts
                start = story_steps[s].create_copy(use_new_id=False)
                start.end_checkpoints = [
                    cp for cp in start.end_checkpoints if cp.name not in overlapping_cps
                ]
                start.end_checkpoints.append(Checkpoint(sink_cp_name))
                story_steps[s] = start

                needs_connector = False

                for k, step in list(story_steps.items()):
                    additional_ends = []
                    for original_cp in overlapping_cps:
                        for cp in step.start_checkpoints:
                            if cp.name == original_cp:
                                if k == e:
                                    cp_name = source_cp_name
                                else:
                                    cp_name = connector_cp_name
                                    needs_connector = True

                                if not self._is_checkpoint_in_list(
                                    cp_name, cp.conditions, step.start_checkpoints
                                ):
                                    # add checkpoint only if it was not added
                                    additional_ends.append(
                                        Checkpoint(cp_name, cp.conditions)
                                    )

                    if additional_ends:
                        updated = step.create_copy(use_new_id=False)
                        updated.start_checkpoints.extend(additional_ends)
                        story_steps[k] = updated

                if needs_connector:
                    start.end_checkpoints.append(Checkpoint(connector_cp_name))

        # the process above may generate unused checkpoints
        # we need to find them and remove them
        self._remove_unused_generated_cps(
            story_steps, all_overlapping_cps, story_end_checkpoints
        )

        return StoryGraph(list(story_steps.values()), story_end_checkpoints)

    @staticmethod
    def _checkpoint_difference(
        cps: List[Checkpoint], cp_name_to_ignore: Set[Text]
    ) -> List[Checkpoint]:
        """Finds checkpoints which names are
            different form names of checkpoints to ignore"""

        return [cp for cp in cps if cp.name not in cp_name_to_ignore]

    def _remove_unused_generated_cps(
        self,
        story_steps: Dict[Text, StoryStep],
        overlapping_cps: Set[Text],
        story_end_checkpoints: Dict[Text, Text],
    ) -> None:
        """Finds unused generated checkpoints
            and remove them from story steps."""

        unused_cps = self._find_unused_checkpoints(
            story_steps.values(), story_end_checkpoints
        )

        unused_overlapping_cps = unused_cps.intersection(overlapping_cps)

        unused_genr_cps = {
            cp_name
            for cp_name in unused_cps
            if cp_name.startswith(GENERATED_CHECKPOINT_PREFIX)
        }

        k_to_remove = set()
        for k, step in story_steps.items():
            # changed all ends
            updated = step.create_copy(use_new_id=False)
            updated.start_checkpoints = self._checkpoint_difference(
                updated.start_checkpoints, unused_overlapping_cps
            )

            # remove generated unused end checkpoints
            updated.end_checkpoints = self._checkpoint_difference(
                updated.end_checkpoints, unused_genr_cps
            )

            if (
                step.start_checkpoints
                and not updated.start_checkpoints
                or step.end_checkpoints
                and not updated.end_checkpoints
            ):
                # remove story step if the generated checkpoints
                # were the only ones
                k_to_remove.add(k)

            story_steps[k] = updated

        # remove unwanted story steps
        for k in k_to_remove:
            del story_steps[k]

    @staticmethod
    def _is_checkpoint_in_list(
        checkpoint_name: Text, conditions: Dict[Text, Any], cps: List[Checkpoint]
    ) -> bool:
        """Checks if checkpoint with name and conditions is
            already in the list of checkpoints."""

        for cp in cps:
            if checkpoint_name == cp.name and conditions == cp.conditions:
                return True
        return False

    @staticmethod
    def _find_unused_checkpoints(
        story_steps: ValuesView[StoryStep], story_end_checkpoints: Dict[Text, Text]
    ) -> Set[Text]:
        """Finds all unused checkpoints."""

        collected_start = {STORY_END, STORY_START}
        collected_end = {STORY_END, STORY_START}

        for step in story_steps:
            for start in step.start_checkpoints:
                collected_start.add(start.name)
            for end in step.end_checkpoints:
                start_name = story_end_checkpoints.get(end.name, end.name)
                collected_end.add(start_name)

        return collected_end.symmetric_difference(collected_start)

    def get(self, step_id: Text) -> Optional[StoryStep]:
        """Looks a story step up by its id."""

        return self.step_lookup.get(step_id)

    def as_story_string(self) -> Text:
        """Convert the graph into the story file format."""

        story_content = ""
        for step in self.story_steps:
            story_content += step.as_story_string(flat=False)
        return story_content

    @staticmethod
    def order_steps(
        story_steps: List[StoryStep],
    ) -> Tuple[deque, List[Tuple[Text, Text]]]:
        """Topological sort of the steps returning the ids of the steps."""

        checkpoints = StoryGraph._group_by_start_checkpoint(story_steps)
        graph = {
            s.id: {
                other.id for end in s.end_checkpoints for other in checkpoints[end.name]
            }
            for s in story_steps
        }
        return StoryGraph.topological_sort(graph)

    @staticmethod
    def _group_by_start_checkpoint(
        story_steps: List[StoryStep],
    ) -> Dict[Text, List[StoryStep]]:
        """Returns all the start checkpoint of the steps"""

        checkpoints = defaultdict(list)
        for step in story_steps:
            for start in step.start_checkpoints:
                checkpoints[start.name].append(step)
        return checkpoints

    @staticmethod
    def topological_sort(
        graph: Dict[Text, Set[Text]]
    ) -> Tuple[deque, List[Tuple[Text, Text]]]:
        """Creates a top sort of a directed graph. This is an unstable sorting!

        The function returns the sorted nodes as well as the edges that need
        to be removed from the graph to make it acyclic (and hence, sortable).

        The graph should be represented as a dictionary, e.g.:

        >>> example_graph = {
        ...         "a": set("b", "c", "d"),
        ...         "b": set(),
        ...         "c": set("d"),
        ...         "d": set(),
        ...         "e": set("f"),
        ...         "f": set()}
        >>> StoryGraph.topological_sort(example_graph)
        (deque([u'e', u'f', u'a', u'c', u'd', u'b']), [])
        """

        # noinspection PyPep8Naming
        GRAY, BLACK = 0, 1

        ordered = deque()
        unprocessed = sorted(set(graph))
        visited_nodes = {}

        removed_edges = set()

        def dfs(node):
            visited_nodes[node] = GRAY
            for k in sorted(graph.get(node, set())):
                sk = visited_nodes.get(k, None)
                if sk == GRAY:
                    removed_edges.add((node, k))
                    continue
                if sk == BLACK:
                    continue
                unprocessed.remove(k)
                dfs(k)
            ordered.appendleft(node)
            visited_nodes[node] = BLACK

        while unprocessed:
            dfs(unprocessed.pop())

        return ordered, sorted(removed_edges)

    def visualize(self, output_file=None):
        import networkx as nx
        from rasa.core.training import visualization  # pytype: disable=pyi-error
        from colorhash import ColorHash

        graph = nx.MultiDiGraph()
        next_node_idx = [0]
        nodes = {"STORY_START": 0, "STORY_END": -1}

        def ensure_checkpoint_is_drawn(cp):
            if cp.name not in nodes:
                next_node_idx[0] += 1
                nodes[cp.name] = next_node_idx[0]

                if cp.name.startswith(GENERATED_CHECKPOINT_PREFIX):
                    # colors generated checkpoints based on their hash
                    color = ColorHash(cp.name[-GENERATED_HASH_LENGTH:]).hex
                    graph.add_node(
                        next_node_idx[0],
                        label=utils.cap_length(cp.name),
                        style="filled",
                        fillcolor=color,
                    )
                else:
                    graph.add_node(next_node_idx[0], label=utils.cap_length(cp.name))

        graph.add_node(
            nodes["STORY_START"], label="START", fillcolor="green", style="filled"
        )
        graph.add_node(nodes["STORY_END"], label="END", fillcolor="red", style="filled")

        for step in self.story_steps:
            next_node_idx[0] += 1
            step_idx = next_node_idx[0]

            graph.add_node(
                next_node_idx[0],
                label=utils.cap_length(step.block_name),
                style="filled",
                fillcolor="lightblue",
                shape="rect",
            )

            for c in step.start_checkpoints:
                ensure_checkpoint_is_drawn(c)
                graph.add_edge(nodes[c.name], step_idx)
            for c in step.end_checkpoints:
                ensure_checkpoint_is_drawn(c)
                graph.add_edge(step_idx, nodes[c.name])

            if not step.end_checkpoints:
                graph.add_edge(step_idx, nodes["STORY_END"])

        if output_file:
            visualization.persist_graph(graph, output_file)

        return graph

    def is_empty(self) -> bool:
        """Checks if `StoryGraph` is empty."""

        return not self.story_steps



import argparse
import logging
from typing import Any, Optional, Text, Tuple, Union

from rasa.nlu import config, utils
from rasa.nlu.components import ComponentBuilder
from rasa.nlu.config import RasaNLUModelConfig
from rasa.nlu.model import Interpreter, Trainer
from rasa.nlu.training_data import load_data
from rasa.nlu.training_data.loading import load_data_from_endpoint
from rasa.nlu.utils import read_endpoints
from rasa.utils.endpoints import EndpointConfig

logger = logging.getLogger(__name__)


def create_argument_parser():
    parser = argparse.ArgumentParser(description="train a custom language parser")

    parser.add_argument(
        "-o",
        "--path",
        default="models/nlu/",
        help="Path where model files will be saved",
    )

    group = parser.add_mutually_exclusive_group(required=True)

    group.add_argument(
        "-d",
        "--data",
        default=None,
        help="Location of the training data. For JSON and "
        "markdown data, this can either be a single file "
        "or a directory containing multiple training "
        "data files.",
    )

    group.add_argument(
        "-u", "--url", default=None, help="URL from which to retrieve training data."
    )

    group.add_argument(
        "--endpoints",
        default=None,
        help="EndpointConfig defining the server from which pull training data.",
    )

    parser.add_argument(
        "-c", "--config", required=True, help="Rasa NLU configuration file"
    )

    parser.add_argument(
        "-t",
        "--num_threads",
        default=1,
        type=int,
        help="Number of threads to use during model training",
    )

    parser.add_argument(
        "--project", default=None, help="Project this model belongs to."
    )

    parser.add_argument(
        "--fixed_model_name",
        help="If present, a model will always be persisted "
        "in the specified directory instead of creating "
        "a folder like 'model_20171020-160213'",
    )

    parser.add_argument(
        "--storage",
        help="Set the remote location where models are stored. "
        "E.g. on AWS. If nothing is configured, the "
        "server will only serve the models that are "
        "on disk in the configured `path`.",
    )

    utils.add_logging_option_arguments(parser)
    return parser


class TrainingException(Exception):
    """Exception wrapping lower level exceptions that may happen while training
      Attributes:
          failed_target_project -- name of the failed project
          message -- explanation of why the request is invalid
      """

    def __init__(self, failed_target_project=None, exception=None):
        self.failed_target_project = failed_target_project
        if exception:
            self.message = exception.args[0]

    def __str__(self):
        return self.message


def create_persistor(persistor: Optional[Text]):
    """Create a remote persistor to store the model if configured."""

    if persistor is not None:
        from rasa.nlu.persistor import get_persistor

        return get_persistor(persistor)
    else:
        return None


def do_train_in_worker(
    cfg: RasaNLUModelConfig,
    data: Text,
    path: Text,
    fixed_model_name: Optional[Text] = None,
    storage: Optional[Text] = None,
    component_builder: Optional[ComponentBuilder] = None,
) -> Text:
    """Loads the trainer and the data and runs the training in a worker."""

    try:
        _, _, persisted_path = train(
            cfg, data, path, fixed_model_name, storage, component_builder
        )
        return persisted_path
    except BaseException as e:
        logger.exception("Failed to train on data '{}'.".format(data))
        raise TrainingException(path, e)


def train(
    nlu_config: Union[Text, RasaNLUModelConfig],
    data: Text,
    path: Optional[Text] = None,
    fixed_model_name: Optional[Text] = None,
    storage: Optional[Text] = None,
    component_builder: Optional[ComponentBuilder] = None,
    training_data_endpoint: Optional[EndpointConfig] = None,
    **kwargs: Any
) -> Tuple[Trainer, Interpreter, Text]:
    """Loads the trainer and the data and runs the training of the model."""

    if isinstance(nlu_config, str):
        nlu_config = config.load(nlu_config)

    # Ensure we are training a model that we can save in the end
    # WARN: there is still a race condition if a model with the same name is
    # trained in another subprocess
    trainer = Trainer(nlu_config, component_builder)
    persistor = create_persistor(storage)
    if training_data_endpoint is not None:
        training_data = load_data_from_endpoint(
            training_data_endpoint, nlu_config.language
        )
    else:
        training_data = load_data(data, nlu_config.language)
    interpreter = trainer.train(training_data, **kwargs)

    if path:
        persisted_path = trainer.persist(path, persistor, fixed_model_name)
    else:
        persisted_path = None

    return trainer, interpreter, persisted_path


if __name__ == "__main__":
    arg_parser = create_argument_parser()
    cmdline_args = arg_parser.parse_args()



import subprocess
import sys
import pytest
from typing import Text

NUMBER_OF_MEASUREMENTS = 15

# Maximum expected import time for rasa module when running on a Travis VM.
# Keep in mind the hardware configuration where tests are run:
# https://docs.travis-ci.com/user/reference/overview/
MAX_IMPORT_TIME_S = 0.3


def _average_import_time(n: int, module: Text) -> float:
    total = 0

    for _ in range(n):
        lines = subprocess.getoutput(
            f'{sys.executable} -X importtime -c "import {module}"'
        ).splitlines()

        parts = lines[-1].split("|")
        if parts[-1].strip() != module:
            raise Exception(f"Import time not found for {module}.")

        total += int(parts[1].strip()) / 1000000

    return total / n


@pytest.mark.skipif(sys.version_info < (3, 7), reason="Need 3.7+ for -X importtime")
def test_import_time():
    import_time = _average_import_time(NUMBER_OF_MEASUREMENTS, "rasa")
    assert import_time < MAX_IMPORT_TIME_S



MESSAGE_TEXT_ATTRIBUTE = "text"

MESSAGE_RESPONSE_KEY_ATTRIBUTE = "response_key"

MESSAGE_INTENT_ATTRIBUTE = "intent"

MESSAGE_RESPONSE_ATTRIBUTE = "response"

MESSAGE_ENTITIES_ATTRIBUTE = "entities"

MESSAGE_NER_FEATURES_ATTRIBUTE = "ner_features"

MESSAGE_ATTRIBUTES = [
    MESSAGE_TEXT_ATTRIBUTE,
    MESSAGE_INTENT_ATTRIBUTE,
    MESSAGE_RESPONSE_ATTRIBUTE,
]

MESSAGE_TOKENS_NAMES = {
    MESSAGE_TEXT_ATTRIBUTE: "tokens",
    MESSAGE_INTENT_ATTRIBUTE: "intent_tokens",
    MESSAGE_RESPONSE_ATTRIBUTE: "response_tokens",
    MESSAGE_ENTITIES_ATTRIBUTE: "tokens",
}

MESSAGE_VECTOR_FEATURE_NAMES = {
    MESSAGE_TEXT_ATTRIBUTE: "text_features",
    MESSAGE_INTENT_ATTRIBUTE: "intent_features",
    MESSAGE_RESPONSE_ATTRIBUTE: "response_features",
    MESSAGE_ENTITIES_ATTRIBUTE: "ner_features",
}

MESSAGE_SPACY_FEATURES_NAMES = {
    MESSAGE_TEXT_ATTRIBUTE: "spacy_doc",
    MESSAGE_RESPONSE_ATTRIBUTE: "response_spacy_doc",
}

SPACY_FEATURIZABLE_ATTRIBUTES = [MESSAGE_TEXT_ATTRIBUTE, MESSAGE_RESPONSE_ATTRIBUTE]

MESSAGE_SELECTOR_PROPERTY_NAME = "response_selector"
DEFAULT_OPEN_UTTERANCE_TYPE = "default"
OPEN_UTTERANCE_PREDICTION_KEY = "response"
OPEN_UTTERANCE_RANKING_KEY = "ranking"
RESPONSE_IDENTIFIER_DELIMITER = "/"



import tempfile
import os
import shutil
from typing import Text

import pytest
from _pytest.monkeypatch import MonkeyPatch
from _pytest.tmpdir import TempdirFactory

import rasa.model

from rasa.train import train_core, train_nlu, train
from tests.core.test_model import _fingerprint


@pytest.mark.parametrize(
    "parameters",
    [
        {"model_name": "test-1234", "prefix": None},
        {"model_name": None, "prefix": "core-"},
        {"model_name": None, "prefix": None},
    ],
)
def test_package_model(trained_rasa_model, parameters):
    output_path = tempfile.mkdtemp()
    train_path = rasa.model.unpack_model(trained_rasa_model)

    model_path = rasa.model.package_model(
        _fingerprint(),
        output_path,
        train_path,
        parameters["model_name"],
        parameters["prefix"],
    )

    assert os.path.exists(model_path)

    file_name = os.path.basename(model_path)

    if parameters["model_name"]:
        assert parameters["model_name"] in file_name

    if parameters["prefix"]:
        assert parameters["prefix"] in file_name

    assert file_name.endswith(".tar.gz")


def count_temp_rasa_files(directory: Text) -> int:
    return len(
        [
            entry
            for entry in os.listdir(directory)
            if not any(
                [
                    # Ignore the following files/directories:
                    entry == "__pycache__",  # Python bytecode
                    entry.endswith(".py")  # Temp .py files created by TF
                    # Anything else is considered to be created by Rasa
                ]
            )
        ]
    )


def test_train_temp_files(
    tmp_path: Text,
    monkeypatch: MonkeyPatch,
    default_domain_path: Text,
    default_stories_file: Text,
    default_stack_config: Text,
    default_nlu_data: Text,
):
    monkeypatch.setattr(tempfile, "tempdir", tmp_path)
    output = "test_train_temp_files_models"

    train(
        default_domain_path,
        default_stack_config,
        [default_stories_file, default_nlu_data],
        output=output,
        force_training=True,
    )

    assert count_temp_rasa_files(tempfile.tempdir) == 0

    # After training the model, try to do it again. This shouldn't try to train
    # a new model because nothing has been changed. It also shouldn't create
    # any temp files.
    train(
        default_domain_path,
        default_stack_config,
        [default_stories_file, default_nlu_data],
        output=output,
    )

    assert count_temp_rasa_files(tempfile.tempdir) == 0


def test_train_core_temp_files(
    tmp_path: Text,
    monkeypatch: MonkeyPatch,
    default_domain_path: Text,
    default_stories_file: Text,
    default_stack_config: Text,
):
    monkeypatch.setattr(tempfile, "tempdir", tmp_path)

    train_core(
        default_domain_path,
        default_stack_config,
        default_stories_file,
        output="test_train_core_temp_files_models",
    )

    assert count_temp_rasa_files(tempfile.tempdir) == 0


def test_train_nlu_temp_files(
    tmp_path: Text,
    monkeypatch: MonkeyPatch,
    default_stack_config: Text,
    default_nlu_data: Text,
):
    monkeypatch.setattr(tempfile, "tempdir", tmp_path)

    train_nlu(
        default_stack_config,
        default_nlu_data,
        output="test_train_nlu_temp_files_models",
    )

    assert count_temp_rasa_files(tempfile.tempdir) == 0
